﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace AirFighter
{
    internal abstract class AbstractMap : IEquatable<AbstractMap>
    {
        private IDrawingObject _drawningObject = null;
        protected int[,] _map = null;
        protected int _width;
        protected int _height;
        protected float _size_x;
        protected float _size_y;
        protected readonly Random _random = new();
        protected readonly int _freeRoad = 0;
        protected readonly int _barrier = 1;
        public Bitmap CreateMap(int width, int height, IDrawingObject drawningObject)
        {
            _width = width;
            _height = height;
            _drawningObject = drawningObject;
            GenerateMap();
            while (!SetObjectOnMap())
            {
                GenerateMap();
            }
            return DrawMapWithObject();
        }

        private (int mapI, int mapJ) checkBarrier((float Left, float Right, float Top, float Bottom) rect)
        {
            return checkBarrier(rect, false, false, true, false);
        }
        private (int mapI, int mapJ) checkBarrier((float Left, float Right, float Top, float Bottom) rect, bool minLeft, bool maxLeft, bool isTop, bool isBottom) 
        {
            (int mapI, int mapJ) res = (-1, -1);

            for (int i = (int)(rect.Top / _size_y); i <= (int)(rect.Bottom / _size_y) && i < _map.GetLength(0); ++i)
            {
                for (int j = (int)(rect.Left / _size_x); j <= (int)(rect.Right / _size_x) && j < _map.GetLength(1); ++j)
                {
                    if (j < 0) j = 0;
                    if (i < 0) i = 0;
                    if (_map[i, j] != _barrier) continue;

                    if (res.mapI == -1) res = (i, j);
                    if (minLeft && res.mapJ > j) res = (i, j);
                    if (maxLeft && res.mapJ < j) res = (i, j);
                    if(isBottom) res = (i, j);
                    if (isTop) return (i, j);
                }
            }

            return res;
        }
        public Bitmap MoveObject(Direction direction)
        {
            // TODO проверка, что объект может переместится в требуемом

            var position = _drawningObject.GetCurrentPosition();

            float drawningWidth = position.Right - position.Left;
            float drawningHeight = position.Bottom - position.Top;

            bool minLeft = false;
            bool maxLeft = false;
            bool isTop = false;
            bool isBottom = false;

            if (direction == Direction.Left)
            {
                position.Left -= _drawningObject.Step;
                maxLeft = true;
            }
            if (direction == Direction.Right)
            {
                position.Left += _drawningObject.Step;
                minLeft = true;
            }
            if (direction == Direction.Up) { 
                position.Top -= _drawningObject.Step;
                isTop = true;
            }    
            if (direction == Direction.Down)
            {
                position.Top += _drawningObject.Step;
                isBottom = true;
            }



            position.Right = position.Left + drawningWidth;
            position.Bottom = position.Top + drawningHeight;

            var currentBarrier = checkBarrier(position, minLeft, maxLeft, isTop, isBottom);

            if (currentBarrier.mapI == -1)
            {
                _drawningObject.MoveObject(direction);
            }

            else if (direction == Direction.Left)
                position.Left = (currentBarrier.mapJ + 1) * _size_x + 1;

            else if (direction == Direction.Right)
                position.Left = currentBarrier.mapJ * _size_x - drawningWidth - 1;

            else if (direction == Direction.Up)
                position.Top = (currentBarrier.mapI + 1) * _size_y + 1;

            else if (direction == Direction.Down)
                position.Top = currentBarrier.mapI * _size_y - drawningHeight - 1;

            if (currentBarrier.mapI != -1)
                _drawningObject.SetObject((int)position.Left, (int)position.Top, _width, _height);

            return DrawMapWithObject();
        }
        private bool SetObjectOnMap()
        {
            if (_drawningObject == null || _map == null)
            {
                return false;
            }
            int x = _random.Next(0, 10);
            int y = _random.Next(0, 10);
            _drawningObject.SetObject(x, y, _width, _height);
            // TODO првоерка, что объект не "накладывается" на закрытые участки

            var position = _drawningObject.GetCurrentPosition();
            float drawningWidth = position.Right - position.Left;
            float drawningHeight = position.Bottom - position.Top;

            var currentBarrier = checkBarrier(position);
            int minRowIndex = _map.GetLength(0);

            while(currentBarrier.mapI != -1)
            {
                minRowIndex = currentBarrier.mapI < minRowIndex ? currentBarrier.mapI : minRowIndex;

                position.Left = (currentBarrier.mapJ + 1) * _size_x + 1;
                position.Right = position.Left + drawningWidth;

                if(position.Right > _width)
                {
                    position.Top = (minRowIndex + 1) * _size_y + 1;
                    position.Bottom = position.Top + drawningHeight;
                     
                    position.Left = 0;
                    position.Right = drawningWidth;

                    minRowIndex = _map.GetLength(0);
                }

                if (position.Bottom > _height) { 
                    return false; 
                }

                currentBarrier = checkBarrier(position);
            }

            _drawningObject.SetObject((int)position.Left, (int)position.Top, _width, _height);

            return true;
        }
        private Bitmap DrawMapWithObject()
        {
            Bitmap bmp = new(_width, _height);
            if (_drawningObject == null || _map == null)
            {
                return bmp;
            }
            Graphics gr = Graphics.FromImage(bmp);
            for (int i = 0; i < _map.GetLength(0); ++i)
            {
                for (int j = 0; j < _map.GetLength(1); ++j)
                {
                    if (_map[i, j] == _freeRoad)
                    {
                        DrawRoadPart(gr, i, j);
                    }
                    else if (_map[i, j] == _barrier)
                    {
                        DrawBarrierPart(gr, i, j);
                    }
                }
            }
            _drawningObject.DrawningObject(gr);
            return bmp;
        }
        protected abstract void GenerateMap();
        protected abstract void DrawRoadPart(Graphics g, int i, int j);
        protected abstract void DrawBarrierPart(Graphics g, int i, int j);

        public bool Equals(AbstractMap? other)
        {
            if (_width != other._width || _height != other._height) return false;
            if (_size_x != other._size_x || _size_y != other._size_y) return false;

            for(int i = 0; i < _map.GetLength(0); i++)
            {
                for(int j = 0; j < _map.GetLength(1); j++)
                {
                    if(_map[i, j] != other._map[i, j]) return false;
                }
            }

            return true;
        }
    }
}
